# Distributed under the OSI-approved BSD 3-Clause License.  See accompanying
# file LICENSE.rst or https://cmake.org/licensing for details.

# BEGIN imports

import os
import re
from dataclasses import dataclass
from typing import Any, List, Tuple, Type, cast

import sphinx

# The following imports may fail if we don't have Sphinx 2.x or later.
if sphinx.version_info >= (2,):
    from docutils import io, nodes
    from docutils.statemachine import ViewList
    from docutils.nodes import Element, Node, TextElement, system_message
    from docutils.parsers.rst import Directive, directives
    from docutils.transforms import Transform
    from docutils.utils.code_analyzer import Lexer, LexerError

    from sphinx import addnodes
    from sphinx.directives import ObjectDescription, nl_escape_re
    from sphinx.domains import Domain, ObjType
    from sphinx.domains.changeset import VersionChange
    from sphinx.domains.changeset import versionlabels, versionlabel_classes
    from sphinx.domains.std import OptionXRefRole
    from sphinx.roles import XRefRole
    from sphinx.util import logging, ws_re
    from sphinx.util.docutils import ReferenceRole
    from sphinx.util.nodes import make_refnode
else:
    # Sphinx 2.x is required.
    assert sphinx.version_info >= (2,)

# END imports

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

# BEGIN sphinx tweaks

# Adjust the 'std' domain regex used to parse options so that it recognizes
# e.g. `-W<name>` as the option `-W` with a value `<name>`, rather than
# treating the entire string as the option name.
#
# See also https://github.com/sphinx-doc/sphinx/issues/14323.

sphinx.domains.std.option_desc_re = (
    re.compile(r'((?:/|--|-|\+)?[^\s=<]+)(=?\s*.*)'))

# END sphinx tweaks

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

# BEGIN pygments tweaks

# Override much of pygments' CMakeLexer.
# We need to parse CMake syntax definitions, not CMake code.

# For hard test cases that use much of the syntax below, see
# - module/FindPkgConfig.html
#     (with "glib-2.0>=2.10 gtk+-2.0" and similar)
# - module/ExternalProject.html
#     (with http:// https:// git@; also has command options -E --build)
# - manual/cmake-buildsystem.7.html
#     (with nested $<..>; relative and absolute paths, "::")

from pygments.lexer import bygroups  # noqa I100
from pygments.lexers import CMakeLexer
from pygments.token import (Comment, Name, Number, Operator, Punctuation,
                            String, Text, Whitespace)

# Notes on regular expressions below:
# - [\.\+-] are needed for string constants like gtk+-2.0
# - Unix paths are recognized by '/'; support for Windows paths may be added
#   if needed
# - (\\.) allows for \-escapes (used in manual/cmake-language.7)
# - $<..$<..$>..> nested occurrence in cmake-buildsystem
# - Nested variable evaluations are only supported in a limited capacity.
#   Only one level of nesting is supported and at most one nested variable can
#   be present.

CMakeLexer.tokens["root"] = [
  # [[string]]
  (r'\[\[(?:\W[\w\W]*?)?\]\]', String.Multiline),
  (r'\[(?P<level>=+)\[[\w\W]*?\](?P=level)\]', String.Multiline),
  # fctn(
  (r'\b(\w+)([ \t]*)(\()',
   bygroups(Name.Function, Text, Name.Function), '#push'),
  (r'\(', Name.Function, '#push'),
  (r'\)', Name.Function, '#pop'),
  (r'\[', Punctuation, '#push'),
  (r'\]', Punctuation, '#pop'),
  (r'[|;,.=*\-]', Punctuation),
  # used in commands/source_group
  (r'\\\\', Punctuation),
  (r'[:]', Operator),
  # used in FindPkgConfig.cmake
  (r'[<>]=', Punctuation),
  # $<...>
  (r'\$<', Operator, '#push'),
  # <expr>
  (r'<[^<|]+?>(\w*\.\.\.)?', Name.Variable),
  # ${..} $ENV{..}, possibly nested
  (r'(\$\w*\{)([^\}\$]*)?(?:(\$\w*\{)([^\}]+?)(\}))?([^\}]*?)(\})',
   bygroups(Operator, Name.Tag, Operator, Name.Tag, Operator, Name.Tag,
            Operator)),
  # DATA{ ...}
  (r'([A-Z]+\{)(.+?)(\})', bygroups(Operator, Name.Tag, Operator)),
  # URL, git@, ...
  (r'[a-z]+(@|(://))((\\.)|[\w.+-:/\\])+', Name.Attribute),
  # absolute path
  (r'/\w[\w\.\+-/\\]*', Name.Attribute),
  (r'/', Name.Attribute),
  # relative path
  (r'\w[\w\.\+-]*/[\w.+-/\\]*', Name.Attribute),
  # initial A-Z, contains a-z
  (r'[A-Z]((\\.)|[\w.+-])*[a-z]((\\.)|[\w.+-])*', Name.Builtin),
  (r'@?[A-Z][A-Z0-9_]*', Name.Constant),
  (r'[a-z_]((\\;)|(\\ )|[\w.+-])*', Name.Builtin),
  (r'[0-9][0-9\.]*', Number),
  # "string"
  (r'(?s)"(\\"|[^"])*"', String),
  (r'\.\.\.', Name.Variable),
  # <..|..> is different from <expr>
  # TODO: <..|..> should be {..|..}; remove when all instances are converted?
  (r'<', Punctuation, '#push'),
  (r'>', Punctuation, '#pop'),
  # {..|..}
  (r'\{', Punctuation, '#push'),
  (r'\}', Punctuation, '#pop'),
  (r'\n', Whitespace),
  (r'[ \t]+', Whitespace),
  (r'#.*\n', Comment),
  # fallback, for debugging only
  #  (r'[^<>\])\}\|$"# \t\n]+', Name.Exception),
]

# END pygments tweaks

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

logger = logging.getLogger(__name__)

# RE to split multiple command signatures.
sig_end_re = re.compile(r'(?<=[)])\n')


def _parse_rst(state, text: str) -> list[Node]:
    node = nodes.Element()
    state.nested_parse(ViewList([text], source=''), 0, node)
    return node.children


@dataclass
class ObjectEntry:
    docname: str
    objtype: str
    node_id: str
    name: str


class CMakeModule(Directive):
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {'encoding': directives.encoding}

    def __init__(self, *args, **keys):
        self.re_start = re.compile(r'^#\[(?P<eq>=*)\[\.rst:$')
        Directive.__init__(self, *args, **keys)

    def run(self):
        settings = self.state.document.settings
        if not settings.file_insertion_enabled:
            raise self.warning(f'{self.name!r} directive disabled.')

        env = self.state.document.settings.env
        rel_path, path = env.relfn2path(self.arguments[0])
        path = os.path.normpath(path)
        encoding = self.options.get('encoding', settings.input_encoding)
        e_handler = settings.input_encoding_error_handler
        try:
            settings.record_dependencies.add(path)
            f = io.FileInput(source_path=path, encoding=encoding,
                             error_handler=e_handler)
        except UnicodeEncodeError:
            msg = (f'Problems with {self.name!r} directive path:\n'
                   f'Cannot encode input file path {path!r} (wrong locale?).')
            raise self.severe(msg)
        except IOError as error:
            msg = f'Problems with {self.name!r} directive path:\n{error}.'
            raise self.severe(msg)
        raw_lines = f.read().splitlines()
        f.close()
        rst = None
        lines = []
        for line in raw_lines:
            if rst is not None and rst != '#':
                # Bracket mode: check for end bracket
                pos = line.find(rst)
                if pos >= 0:
                    if line[0] == '#':
                        line = ''
                    else:
                        line = line[0:pos]
                    rst = None
            else:
                # Line mode: check for .rst start (bracket or line)
                m = self.re_start.match(line)
                if m:
                    rst = f']{m.group("eq")}]'
                    line = ''
                elif line == '#.rst:':
                    rst = '#'
                    line = ''
                elif rst == '#':
                    if line == '#' or line[:2] == '# ':
                        line = line[2:]
                    else:
                        rst = None
                        line = ''
                elif rst is None:
                    line = ''
            lines.append(line)
        if rst is not None and rst != '#':
            raise self.warning(f'{self.name!r} found unclosed bracket '
                               f'"#[{rst[1:-1]}[.rst:" in {path!r}')
        self.state_machine.insert_input(lines, path)
        return []


class _cmake_index_entry:
    def __init__(self, desc):
        self.desc = desc

    def __call__(self, title, targetid, main='main'):
        return ('pair', f'{self.desc} ; {title}', targetid, main, None)


_cmake_index_objs = {
    'command':    _cmake_index_entry('command'),
    'cpack_gen':  _cmake_index_entry('cpack generator'),
    'diagnostic': _cmake_index_entry('diagnostic'),
    'envvar':     _cmake_index_entry('envvar'),
    'generator':  _cmake_index_entry('generator'),
    'genex':      _cmake_index_entry('genex'),
    'guide':      _cmake_index_entry('guide'),
    'manual':     _cmake_index_entry('manual'),
    'module':     _cmake_index_entry('module'),
    'policy':     _cmake_index_entry('policy'),
    'prop_cache': _cmake_index_entry('cache property'),
    'prop_dir':   _cmake_index_entry('directory property'),
    'prop_fs':    _cmake_index_entry('file set property'),
    'prop_gbl':   _cmake_index_entry('global property'),
    'prop_inst':  _cmake_index_entry('installed file property'),
    'prop_sf':    _cmake_index_entry('source file property'),
    'prop_test':  _cmake_index_entry('test property'),
    'prop_tgt':   _cmake_index_entry('target property'),
    'variable':   _cmake_index_entry('variable'),
    }


class CMakeTransform(Transform):

    # Run this transform early since we insert nodes we want
    # treated as if they were written in the documents.
    default_priority = 210

    def __init__(self, document, startnode):
        Transform.__init__(self, document, startnode)
        self.titles = {}

    def parse_title(self, docname):
        """Parse a document title as the first line starting in [A-Za-z0-9<$]
           or fall back to the document basename if no such line exists.
           The cmake --help-*-list commands also depend on this convention.
           Return the title or False if the document file does not exist.
        """
        settings = self.document.settings
        env = settings.env
        title = self.titles.get(docname)
        if title is None:
            fname = os.path.join(env.srcdir, docname+'.rst')
            try:
                f = open(fname, 'r', encoding=settings.input_encoding)
            except IOError:
                title = False
            else:
                for line in f:
                    if len(line) > 0 and (line[0].isalnum() or
                                          line[0] == '<' or line[0] == '$'):
                        title = line.rstrip()
                        break
                f.close()
                if title is None:
                    title = os.path.basename(docname)
            self.titles[docname] = title
        return title

    def apply(self):
        env = self.document.settings.env

        # Treat some documents as cmake domain objects.
        objtype, sep, tail = env.docname.partition('/')
        make_index_entry = _cmake_index_objs.get(objtype)
        if make_index_entry:
            title = self.parse_title(env.docname)
            # Insert the object link target.
            if objtype == 'command':
                targetname = title.lower()
            elif objtype == 'guide' and not tail.endswith('/index'):
                targetname = tail
            else:
                if objtype == 'genex':
                    m = CMakeXRefRole._re_genex.match(title)
                    if m:
                        title = m.group(1)
                targetname = title
            targetid = f'{objtype}:{targetname}'
            targetnode = nodes.target('', '', ids=[targetid])
            self.document.note_explicit_target(targetnode)
            self.document.insert(0, targetnode)
            # Insert the object index entry.
            indexnode = addnodes.index()
            indexnode['entries'] = [make_index_entry(title, targetid)]
            self.document.insert(0, indexnode)

            # Add to cmake domain object inventory
            domain = cast(CMakeDomain, env.get_domain('cmake'))
            domain.note_object(objtype, targetname, targetid, targetid)


class CMakeObject(ObjectDescription):
    def __init__(self, *args, **kwargs):
        self.targetname = None
        super().__init__(*args, **kwargs)

    def handle_signature(self, sig, signode):
        # called from sphinx.directives.ObjectDescription.run()
        signode += addnodes.desc_name(sig, sig)
        return sig

    def add_target_and_index(self, name, sig, signode):
        if self.objtype == 'command':
            targetname = name.lower()
        elif self.targetname:
            targetname = self.targetname
        else:
            targetname = name
        targetid = f'{self.objtype}:{targetname}'
        if targetid not in self.state.document.ids:
            signode['names'].append(targetid)
            signode['ids'].append(targetid)
            signode['first'] = (not self.names)
            self.state.document.note_explicit_target(signode)

            domain = cast(CMakeDomain, self.env.get_domain('cmake'))
            domain.note_object(self.objtype, targetname, targetid, targetid,
                               location=signode)

        make_index_entry = _cmake_index_objs.get(self.objtype)
        if make_index_entry:
            self.indexnode['entries'].append(make_index_entry(name, targetid))


class CMakeGenexObject(CMakeObject):
    option_spec = {
        'target': directives.unchanged,
    }

    def handle_signature(self, sig, signode):
        name = super().handle_signature(sig, signode)

        m = CMakeXRefRole._re_genex.match(sig)
        if m:
            name = m.group(1)

        return name

    def run(self):
        target = self.options.get('target')
        if target is not None:
            self.targetname = target

        return super().run()


class CMakeSignatureObject(CMakeObject):
    object_type = 'signature'

    BREAK_ALL = 'all'
    BREAK_SMART = 'smart'
    BREAK_VERBATIM = 'verbatim'

    BREAK_CHOICES = {BREAK_ALL, BREAK_SMART, BREAK_VERBATIM}

    def break_option(argument):
        return directives.choice(argument, CMakeSignatureObject.BREAK_CHOICES)

    option_spec = {
        'target': directives.unchanged,
        'break': break_option,
    }

    def _break_signature_all(sig: str) -> str:
        return ws_re.sub(' ', sig)

    def _break_signature_verbatim(sig: str) -> str:
        lines = [ws_re.sub('\xa0', line.strip()) for line in sig.split('\n')]
        return ' '.join(lines)

    def _break_signature_smart(sig: str) -> str:
        tokens = []
        for line in sig.split('\n'):
            token = ''
            delim = ''

            for c in line.strip():
                if len(delim) == 0 and ws_re.match(c):
                    if len(token):
                        tokens.append(ws_re.sub('\xa0', token))
                        token = ''
                else:
                    if c == '[':
                        delim += ']'
                    elif c == '<':
                        delim += '>'
                    elif len(delim) and c == delim[-1]:
                        delim = delim[:-1]
                    token += c

            if len(token):
                tokens.append(ws_re.sub('\xa0', token))

        return ' '.join(tokens)

    def __init__(self, *args, **kwargs):
        self.targetnames = {}
        self.break_style = CMakeSignatureObject.BREAK_SMART
        super().__init__(*args, **kwargs)

    def get_signatures(self) -> List[str]:
        content = nl_escape_re.sub('', self.arguments[0])
        lines = sig_end_re.split(content)

        if self.break_style == CMakeSignatureObject.BREAK_VERBATIM:
            fixup = CMakeSignatureObject._break_signature_verbatim
        elif self.break_style == CMakeSignatureObject.BREAK_SMART:
            fixup = CMakeSignatureObject._break_signature_smart
        else:
            fixup = CMakeSignatureObject._break_signature_all

        return [fixup(line.strip()) for line in lines]

    def handle_signature(self, sig, signode):
        language = 'cmake'
        classes = ['code', 'cmake', 'highlight']

        node = addnodes.desc_name(sig, '', classes=classes)

        try:
            tokens = Lexer(sig, language, 'short')
        except LexerError as error:
            if self.state.document.settings.report_level > 2:
                # Silently insert without syntax highlighting.
                tokens = Lexer(sig, language, 'none')
            else:
                raise self.warning(error)

        for classes, value in tokens:
            if value == '\xa0':
                node += nodes.inline(value, value, classes=['nbsp'])
            elif classes:
                node += nodes.inline(value, value, classes=classes)
            else:
                node += nodes.Text(value)

        signode.clear()
        signode += node

        return sig

    def add_target_and_index(self, name, sig, signode):
        sig = sig.replace('\xa0', ' ')
        if sig in self.targetnames:
            sigargs = self.targetnames[sig]
        else:
            def extract_keywords(params):
                if params[-1].endswith(')'):
                    params[-1] = params[-1][:-1]
                for p in params:
                    if p[0].isalpha():
                        yield p
                    else:
                        return

            keywords = extract_keywords(sig.split('(')[1].split())
            sigargs = ' '.join(keywords)
        targetname = sigargs.lower()
        targetid = nodes.make_id(targetname)

        if targetid not in self.state.document.ids:
            signode['names'].append(targetname)
            signode['ids'].append(targetid)
            signode['first'] = (not self.names)
            self.state.document.note_explicit_target(signode)

            # Register the signature as a command object.
            command = sig.split('(')[0].lower()
            refname = f'{command}({sigargs})'
            refid = f'command:{command}({targetname})'

            domain = cast(CMakeDomain, self.env.get_domain('cmake'))
            domain.note_object('command', name=refname, target_id=refid,
                               node_id=targetid, location=signode)

    def run(self):
        self.break_style = CMakeSignatureObject.BREAK_ALL

        targets = self.options.get('target')
        if targets is not None:
            signatures = self.get_signatures()
            targets = [t.strip() for t in targets.split('\n')]
            for signature, target in zip(signatures, targets):
                self.targetnames[signature] = target

        self.break_style = (
            self.options.get('break', CMakeSignatureObject.BREAK_SMART))

        return super().run()


class CMakeDiagnosticObject(CMakeObject):
    object_type = 'diagnostic'
    required_arguments = 0
    optional_arguments = 0

    DEFAULT_CHOICES = {'ignore', 'warn', 'error'}

    def default_option(argument):
        return directives.choice(
            argument, CMakeDiagnosticObject.DEFAULT_CHOICES)

    option_spec = {
        'default': default_option,
        'parent': directives.unchanged,
    }

    def __init__(self, *args, **kwargs):
        self.targetname = None
        super().__init__(*args, **kwargs)

    def _preset_name(self) -> str:
        sep = False
        out = ''
        for c in self.targetname[4:]:
            if c == '_':
                sep = True
            elif sep:
                out += c
                sep = False
            else:
                out += c.lower()
        return out

    def _build_field(self, name: str, content: str | list[Node]) -> Node:
        if type(content) is not list:
            content = _parse_rst(self.state,content)

        name_node = nodes.field_name(text=name)
        body_node = nodes.field_body('', *content)
        return nodes.field('', name_node, body_node)

    def _build_cli(self) -> list[Node]:
        cname = self.targetname[4:].lower().replace('_', '-')
        ctext = f'-W[no-][error=]{cname}'
        return _parse_rst(self.state,f':option:`{ctext} <cmake -W>`')

    def _build_preset_refs(self) -> list[Node]:
        p = self._preset_name()
        w = f':preset:`warnings.{p} <configurePresets.warnings.{p}>`'
        e = f':preset:`errors.{p} <configurePresets.errors.{p}>`'
        return _parse_rst(self.state,f'{w}, {e}')

    def run(self) -> list[Node]:
        self.domain, self.objtype = self.name.split(':', 1)
        doc = self.state.document
        self.targetname = doc.next_node(nodes.title).astext()

        default = self.options['default'].capitalize()
        parent = self.options.get('parent')

        headers = nodes.field_list()
        headers += self._build_field('Command Line', self._build_cli())
        headers += self._build_field('Presets', self._build_preset_refs())
        headers += self._build_field('Default', default)

        if parent:
            parentRef = _parse_rst(self.state,f':diagnostic:`{parent}`')
            headers += self._build_field('Parent', parentRef)

        node = nodes.Element()
        self.state.nested_parse(self.content, self.content_offset, node)
        return [headers] + node.children


class CMakeReferenceRole:
    # See sphinx.util.nodes.explicit_title_re; \x00 escapes '<'.
    _re = re.compile(r'^(.+?)(\s*)(?<!\x00)<(.*?)>$', re.DOTALL)

    @staticmethod
    def _escape_angle_brackets(text: str) -> str:
        # CMake cross-reference targets frequently contain '<' so escape
        # any explicit `<target>` with '<' not preceded by whitespace.
        while True:
            m = CMakeReferenceRole._re.match(text)
            if m and len(m.group(2)) == 0:
                text = f'{m.group(1)}\x00<{m.group(3)}>'
            else:
                break
        return text

    def __class_getitem__(cls, parent: Any):
        class Class(parent):
            def __call__(self, name: str, rawtext: str, text: str,
                         *args, **kwargs
                         ) -> Tuple[List[Node], List[system_message]]:
                text = CMakeReferenceRole._escape_angle_brackets(text)
                return super().__call__(name, rawtext, text, *args, **kwargs)
        return Class


class CMakeCRefRole(CMakeReferenceRole[ReferenceRole]):
    nodeclass: Type[Element] = nodes.reference
    innernodeclass: Type[TextElement] = nodes.literal
    classes: List[str] = ['cmake', 'literal']

    def run(self) -> Tuple[List[Node], List[system_message]]:
        refnode = self.nodeclass(self.rawtext)
        self.set_source_info(refnode)

        refnode['refid'] = nodes.make_id(self.target)
        refnode += self.innernodeclass(self.rawtext, self.title,
                                       classes=self.classes)

        return [refnode], []


class CMakeXRefRole(CMakeReferenceRole[XRefRole]):

    _re_sub = re.compile(r'^([^()\s]+)\s*\(([^()]*)\)$', re.DOTALL)
    _re_genex = re.compile(r'^\$<([^<>:]+)(:[^<>]+)?>$', re.DOTALL)
    _re_guide = re.compile(r'^([^<>/]+)/([^<>]*)$', re.DOTALL)
    _re_explicit = re.compile(r'^([^<>]+)\s+<([^<>]+)>$', re.DOTALL)

    def __call__(self, typ, rawtext, text, *args, **kwargs):
        if typ == 'cmake:command':
            # Translate a CMake command cross-reference of the form:
            #  `command_name(SUB_COMMAND)`
            # to be its own explicit target:
            #  `command_name(SUB_COMMAND) <command_name(SUB_COMMAND)>`
            # so the XRefRole `fix_parens` option does not add more `()`.
            m = CMakeXRefRole._re_sub.match(text)
            if m:
                text = f'{text} <{text}>'
        elif typ == 'cmake:genex':
            m = CMakeXRefRole._re_genex.match(text)
            if m:
                text = f'{text} <{m.group(1)}>'
        elif typ == 'cmake:guide':
            m = CMakeXRefRole._re_guide.match(text)
            if m:
                text = f'{m.group(2)} <{text}>'
        elif typ == 'cmake:preset':
            m = CMakeXRefRole._re_explicit.match(text)
            if m:
                rawtext = f'{m.group(1)} <CMakePresets.{m.group(2)}>'
                text = f'{m.group(1)} <CMakePresets.{m.group(2)}>'
            else:
                text = f'CMakePresets.{text}'
        return super().__call__(typ, rawtext, text, *args, **kwargs)

    # We cannot insert index nodes using the result_nodes method
    # because CMakeXRefRole is processed before substitution_reference
    # nodes are evaluated so target nodes (with 'ids' fields) would be
    # duplicated in each evaluated substitution replacement.  The
    # docutils substitution transform does not allow this.  Instead we
    # use our own CMakeXRefTransform below to add index entries after
    # substitutions are completed.
    #
    # def result_nodes(self, document, env, node, is_ref):
    #     pass


class CMakeOptionXRefRole(OptionXRefRole):
    def __init__(self, command: str) -> None:
        self.command = command
        super().__init__()

    def __call__(self, typ, rawtext, text, *args, **kwargs):
        content = f'{text} <{self.command} {re.split(r"[ =]", text)[0]}>'
        return super().__call__('std:option', text, content, *args, **kwargs)


class CMakeXRefTransform(Transform):

    # Run this transform early since we insert nodes we want
    # treated as if they were written in the documents, but
    # after the sphinx (210) and docutils (220) substitutions.
    default_priority = 221

    # This helper supports docutils < 0.18, which is missing 'findall',
    # and docutils == 0.18.0, which is missing 'traverse'.
    def _document_findall_as_list(self, condition):
        if hasattr(self.document, 'findall'):
            # Fully iterate into a list so the caller can grow 'self.document'
            # while iterating.
            return list(self.document.findall(condition))

        # Fallback to 'traverse' on old docutils, which returns a list.
        return self.document.traverse(condition)

    def apply(self):
        env = self.document.settings.env

        # Find CMake cross-reference nodes and add index and target
        # nodes for them.
        for ref in self._document_findall_as_list(addnodes.pending_xref):
            if not ref['refdomain'] == 'cmake':
                continue

            objtype = ref['reftype']
            make_index_entry = _cmake_index_objs.get(objtype)
            if not make_index_entry:
                continue

            objname = ref['reftarget']
            if objtype == 'guide' and CMakeXRefRole._re_guide.match(objname):
                # Do not index cross-references to guide sections.
                continue

            if objtype == 'command':
                # Index signature references to their parent command.
                objname = objname.split('(')[0].lower()

            targetnum = env.new_serialno(f'index-{objtype}:{objname}')

            targetid = f'index-{targetnum}-{objtype}:{objname}'
            targetnode = nodes.target('', '', ids=[targetid])
            self.document.note_explicit_target(targetnode)

            indexnode = addnodes.index()
            indexnode['entries'] = [make_index_entry(objname, targetid, '')]
            ref.replace_self([indexnode, targetnode, ref])


class CMakeDomain(Domain):
    """CMake domain."""
    name = 'cmake'
    label = 'CMake'
    object_types = {
        'command':    ObjType('command',    'command'),
        'cpack_gen':  ObjType('cpack_gen',  'cpack_gen'),
        'diagnostic': ObjType('diagnostic', 'diagnostic'),
        'envvar':     ObjType('envvar',     'envvar'),
        'generator':  ObjType('generator',  'generator'),
        'genex':      ObjType('genex',      'genex'),
        'guide':      ObjType('guide',      'guide'),
        'variable':   ObjType('variable',   'variable'),
        'module':     ObjType('module',     'module'),
        'policy':     ObjType('policy',     'policy'),
        'prop_cache': ObjType('prop_cache', 'prop_cache'),
        'prop_dir':   ObjType('prop_dir',   'prop_dir'),
        'prop_fs':    ObjType('prop_fs',    'prop_fs'),
        'prop_gbl':   ObjType('prop_gbl',   'prop_gbl'),
        'prop_inst':  ObjType('prop_inst',  'prop_inst'),
        'prop_sf':    ObjType('prop_sf',    'prop_sf'),
        'prop_test':  ObjType('prop_test',  'prop_test'),
        'prop_tgt':   ObjType('prop_tgt',   'prop_tgt'),
        'manual':     ObjType('manual',     'manual'),
    }
    directives = {
        'command':    CMakeObject,
        'envvar':     CMakeObject,
        'diagnostic': CMakeDiagnosticObject,
        'genex':      CMakeGenexObject,
        'signature':  CMakeSignatureObject,
        'variable':   CMakeObject,
        # Other `object_types` cannot be created except by the `CMakeTransform`
    }
    roles = {
        # General CMake reference roles.
        'cref':       CMakeCRefRole(),
        'command':    CMakeXRefRole(fix_parens=True, lowercase=True),
        'cpack_gen':  CMakeXRefRole(),
        'diagnostic': CMakeXRefRole(),
        'envvar':     CMakeXRefRole(),
        'generator':  CMakeXRefRole(),
        'genex':      CMakeXRefRole(),
        'guide':      CMakeXRefRole(),
        'variable':   CMakeXRefRole(),
        'module':     CMakeXRefRole(),
        'policy':     CMakeXRefRole(),
        'prop_cache': CMakeXRefRole(),
        'prop_dir':   CMakeXRefRole(),
        'prop_fs':    CMakeXRefRole(),
        'prop_gbl':   CMakeXRefRole(),
        'prop_inst':  CMakeXRefRole(),
        'prop_sf':    CMakeXRefRole(),
        'prop_test':  CMakeXRefRole(),
        'prop_tgt':   CMakeXRefRole(),
        'manual':     CMakeXRefRole(),
        'preset':     CMakeXRefRole(lowercase=True, warn_dangling=True),
        # Roles for program-specific command-line options without the program
        # name (which add the name to form the ref).
        'cmake-option':           CMakeOptionXRefRole('cmake'),
        'cmake-build-option':     CMakeOptionXRefRole('cmake--build'),
        'cmake-install-option':   CMakeOptionXRefRole('cmake--install'),
        'cmake-workflow-option':  CMakeOptionXRefRole('cmake--workflow'),
        'cpack-option':           CMakeOptionXRefRole('cpack'),
        'ctest-dashboard-option': CMakeOptionXRefRole('ctest-dashboard'),
        'ctest-option':           CMakeOptionXRefRole('ctest'),
    }
    initial_data = {
        'objects': {},  # fullname -> ObjectEntry
    }

    def clear_doc(self, docname):
        to_clear = set()
        for fullname, obj in self.data['objects'].items():
            if obj.docname == docname:
                to_clear.add(fullname)
        for fullname in to_clear:
            del self.data['objects'][fullname]

    def merge_domaindata(self, docnames, otherdata):
        """Merge domaindata from the workers/chunks when they return.

        Called once per parallelization chunk.
        Only used when sphinx is run in parallel mode.

        :param docnames: a Set of the docnames that are part of the current
                         chunk to merge
        :param otherdata: the partial data calculated by the current chunk
        """
        for refname, obj in otherdata['objects'].items():
            if obj.docname in docnames:
                self.data['objects'][refname] = obj

    def resolve_xref(self, env, fromdocname, builder,
                     typ, target, node, contnode):
        targetid = f'{typ}:{target}'
        obj = self.data['objects'].get(targetid)

        if obj is None:
            if typ == 'command':
                # If 'command(args)' wasn't found, try just 'command'.
                # TODO: remove this fallback? warn?
                # logger.warning(f'no match for {targetid}')
                command = target.split('(')[0]
                targetid = f'{typ}:{command}'
                obj = self.data['objects'].get(targetid)
            elif typ == 'preset':
                # Preset references are really just references to plain old
                # explicit targets.
                labels = env.get_domain('std').labels
                docname, labelid, sectname = labels.get(target, ('', '', ''))

                if not docname:
                    return None

                if node['refexplicit']:
                    sectname = node.astext()

                return make_refnode(builder, fromdocname, docname, labelid,
                                    nodes.literal('', sectname), target)

        if obj is None:
            # TODO: warn somehow?
            return None

        return make_refnode(builder, fromdocname, obj.docname, obj.node_id,
                            contnode, target)

    def note_object(self, objtype: str, name: str, target_id: str,
                    node_id: str, location: Any = None):
        if target_id in self.data['objects']:
            other = self.data['objects'][target_id].docname
            logger.warning(
                f'CMake object {target_id!r} also described in {other!r}',
                location=location)

        self.data['objects'][target_id] = ObjectEntry(
            self.env.docname, objtype, node_id, name)

    def get_objects(self):
        for refname, obj in self.data['objects'].items():
            yield (refname, refname, obj.objtype, obj.docname, obj.node_id, 1)


def setup(app):
    app.add_directive('cmake-module', CMakeModule)
    app.add_transform(CMakeTransform)
    app.add_transform(CMakeXRefTransform)
    app.add_domain(CMakeDomain)

    version_directives = {
        'cmakefiles': 'cmakeFiles',
        'codemodel': 'codemodel',
        'presets': 'presets',
        'toolchains': 'toolchains',
    }
    for directive, name in version_directives.items():
        versionlabels.update({
            f'{directive}-versionadded':   f'Added in {name} version %s',
            f'{directive}-versionchanged': f'Changed in {name} version %s',
            f'{directive}-versionremoved': f'Removed in {name} version %s',
        })
        versionlabel_classes.update({
            f'{directive}-versionadded':   'added',
            f'{directive}-versionchanged': 'changed',
            f'{directive}-versionremoved': 'removed',
        })
        app.add_directive(f'{directive}-versionadded', VersionChange)
        app.add_directive(f'{directive}-versionchanged', VersionChange)
        app.add_directive(f'{directive}-versionremoved', VersionChange)

    return {"parallel_read_safe": True}



#!/usr/bin/env python3
"""Generate CMake presets documentation (reST) and JSON schema.

The documentation and JSON schema for CMake presets are both generated from
Help/manual/presets/schema.yaml, which describes the schema in a way that
allows both documentation forms to be combined in a single file, and also makes
it much easier to manage version revisions (compared to editing the JSON schema
'by hand').

All input and output files are expected to be in known locations relative to
this script.

Usage: python3 regenerate-presets.py
"""

import json
import re
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Self

import yaml

LATEST = 1
TYPES = {}
REFS = {}

PRESETS_REL = 'Help/manual/presets'
SCHEMA_YAML_FILENAME = 'schema.yaml'
SCHEMA_JSON_FILENAME = 'schema.json'

WORKSPACE = Path(__file__).parent.parent.parent.absolute()
SCRIPT_REL = Path(__file__).relative_to(WORKSPACE)
PRESETS = WORKSPACE / PRESETS_REL
DIAGNOSTICS = WORKSPACE / 'Source/cmDiagnostics.h'

RST_BANNER = f"""\
.. This file was generated by {SCRIPT_REL.as_posix()}
   from {PRESETS_REL}/{SCHEMA_YAML_FILENAME}. Do not edit.
"""

DIAGNOSTIC_TABLE_MACRO = 'CM_FOR_EACH_DIAGNOSTIC_TABLE'
CONFIGURE_PRESET_PROPERTIES_PATH = (
    'properties', 'configurePresets', 'items', 'properties'
)
WARNING_DESCRIPTION = (
    'An optional boolean. '
    'Equivalent to passing -W{c} or -Wno-{c} on the command line. '
    'This may not be set to false if errors.{p} is set to true.'
)
WARNING_SPHINX_DESCRIPTION = """
An optional boolean. Equivalent to passing :option:`-W{c} <cmake -W>` or
:option:`-Wno-{c} <cmake -Wno->` on the command line.
This may not be set to ``false`` if ``errors.{p}`` is set to ``true``.
""".strip()
ERROR_DESCRIPTION = (
    'An optional boolean. '
    'Equivalent to passing -Werror={c} or -Wno-error={c} on the command line. '
    'This may not be set to true if warnings.{p} is set to false.'
)
ERROR_SPHINX_DESCRIPTION = """
An optional boolean. Equivalent to passing :cmake-option:`-Werror={c}` or
:cmake-option:`-Wno-error={c}` on the command line.
This may not be set to ``true`` if ``warnings.{p}`` is set to ``false``.
""".strip()


# =============================================================================
@dataclass
class Diagnostic:
    presetName: str
    cliName: str
    since: int

    # -------------------------------------------------------------------------
    def format(self, template: str) -> str:
        return template.format(c=self.cliName, p=self.presetName)


# =============================================================================
class Value:

    # -------------------------------------------------------------------------
    def __init__(self, data: list[dict[str, Any]]):
        self.data = data
        self.spans = None

    # -------------------------------------------------------------------------
    @property
    def objects(self) -> dict[str, 'Object']:
        return {}

    # -------------------------------------------------------------------------
    def updateSpans(self, since: int, until: int) -> None:
        self.spans = [(since, until)]

    # -------------------------------------------------------------------------
    def updateRefs(self, objects: dict[str, 'Object']) -> None:
        pass

    # -------------------------------------------------------------------------
    def defs(self) -> dict[str, Any]:
        return {}

    # -------------------------------------------------------------------------
    def schema(self, version: int) -> dict[str, Any]:
        return deepcopy(self.data)


# =============================================================================
class Object(Value):

    # -------------------------------------------------------------------------
    def __init__(self, data: dict[str, Any], name: str,
                 label: str | None = None):
        # Prevent data sharing from modifying unrelated parts of the schema as
        # we pick apart our section.
        data = deepcopy(data)

        self.name = data.pop('id', name)
        self.label = label

        self.properties = {}
        for n, p in data.pop('properties').items():
            pl = '.'.join([label, n]) if label is not None else n
            self.properties[n] = Property(p, n, pl)
        if len(self.properties):
            self.properties[CommentProperty.name] = CommentProperty()

        ap = data.pop('additionalProperties', None)
        if ap is not None:
            self.additionalProperties = buildValue(ap, self.name, self.label)
        else:
            self.additionalProperties = None

        super().__init__(data)

    # -------------------------------------------------------------------------
    @property
    def objects(self) -> dict[str, Self]:
        out = {self.name: self}
        for p in self.properties.values():
            out.update(p.objects)
        if self.additionalProperties is not None:
            out.update(self.additionalProperties.objects)
        return out

    # -------------------------------------------------------------------------
    def updateSpans(self, since: int, until: int) -> None:
        spans = [[(since, until)]]
        for p in self.properties.values():
            p.updateSpans(since, until)
            spans.append(p.spans)
        if self.additionalProperties is not None:
            self.additionalProperties.updateSpans(since, until)
            spans.append(self.additionalProperties.spans)
        self.spans = combineSpans(*spans)

    # -------------------------------------------------------------------------
    def updateRefs(self, objects: dict[str, 'Object']) -> None:
        for p in self.properties.values():
            p.updateRefs(objects)
        if self.additionalProperties is not None:
            self.additionalProperties.updateRefs(objects)

    # -------------------------------------------------------------------------
    def doc(self, since: int, until: int) -> str | None:
        since = max(since, self.spans[0][0])
        until = min(until, self.spans[-1][1])

        out = []
        for _, p in self.properties.items():
            content = p.doc(since, until)
            if not content:
                continue
            out.append(content)

        return '\n\n'.join(out) if len(out) else None

    # -------------------------------------------------------------------------
    def defs(self, exclude: set[str] = {}) -> dict[str, Any]:
        out = {}
        for n, p in self.properties.items():
            if n in exclude:
                continue
            for s in p.spans:
                r = f'{p.label}@{spanRef(*s)}'
                out[r] = p.schema(s[0])
            out.update(p.defs())
        if self.additionalProperties is not None:
            out.update(self.additionalProperties.defs())
        return out

    # -------------------------------------------------------------------------
    def schema(self, version: int, inline: set[str] = {}) -> dict[str, Any]:
        out = deepcopy(self.data)

        props = {}
        required = []
        for n, p in self.properties.items():
            if n in inline:
                props[n] = p.schema(version)
            else:
                r = p.ref(version)
                if r is not None:
                    props[n] = r
            if p.required:
                required.append(n)

        out['properties'] = props
        if len(required):
            out['required'] = required

        if self.additionalProperties is not None:
            ap = self.additionalProperties.schema(version)
            out['additionalProperties'] = ap
        return out


# =============================================================================
class Ref(Value):

    # -------------------------------------------------------------------------
    def __init__(self, data: list[dict[str, Any]]):
        self.target = data.pop('target')
        self.targetSpans = None
        super().__init__(data)

    # -------------------------------------------------------------------------
    def updateSpans(self, since: int, until: int) -> None:
        if self.spans is None:
            super().updateSpans(since, until)
        self.spans = trimSpans(self.spans, since, until)

    # -------------------------------------------------------------------------
    def updateRefs(self, objects: dict[str, 'Object']) -> None:
        global REFS

        t = objects.get(self.target)
        if t is not None:
            since = self.spans[0][0]
            until = self.spans[-1][1]
            self.spans = trimSpans(t.spans, since, until)
            self.targetSpans = t.spans

        rs = REFS.get(self.target, (since, until))
        REFS[self.target] = (min(since, rs[0]), max(until, rs[1]))

    # -------------------------------------------------------------------------
    def schema(self, version: int) -> dict[str, Any]:
        ts = trimSpans(self.targetSpans, *REFS[self.target])
        s = findSpan(ts, version)
        if s is not None:
            return {
                '$ref': f'#/definitions/{self.target}@{spanRef(*s)}',
            }
        return {}


# =============================================================================
class Array(Value):

    # -------------------------------------------------------------------------
    def __init__(self, data: dict[str, Any], name: str, label: str):
        item = data.pop('items')
        self.value = buildValue(item, name, label)

        super().__init__(data)

    # -------------------------------------------------------------------------
    @property
    def objects(self) -> dict[str, Object]:
        return self.value.objects

    # -------------------------------------------------------------------------
    def updateSpans(self, since: int, until: int) -> None:
        self.value.updateSpans(since, until)
        self.spans = self.value.spans

    # -------------------------------------------------------------------------
    def updateRefs(self, objects: dict[str, 'Object']) -> None:
        self.value.updateRefs(objects)

    # -------------------------------------------------------------------------
    def defs(self) -> dict[str, Any]:
        return self.value.defs()

    # -------------------------------------------------------------------------
    def schema(self, version: int) -> dict[str, Any]:
        out = deepcopy(self.data)
        out['items'] = self.value.schema(version)
        return out


# =============================================================================
class Union(Value):

    # -------------------------------------------------------------------------
    def __init__(self, data: list[dict[str, Any]], name: str, label: str):
        self.items = []
        for i, d in enumerate(data):
            self.items.append(buildValue(d, name, self.itemLabel(label, i)))

        super().__init__({})

    # -------------------------------------------------------------------------
    def itemLabel(self, label: str, index: int) -> str:
        return label

    # -------------------------------------------------------------------------
    @property
    def objects(self) -> dict[str, Object]:
        out = {}
        for v in self.items:
            out.update(v.objects)
        return out

    # -------------------------------------------------------------------------
    def updateSpans(self, since: int, until: int) -> None:
        spans = [[(since, until)]]
        for i in self.items:
            i.updateSpans(since, until)
            spans.append(i.spans)
        self.spans = combineSpans(*spans)

    # -------------------------------------------------------------------------
    def updateRefs(self, objects: dict[str, 'Object']) -> None:
        for i in self.items:
            i.updateRefs(objects)

    # -------------------------------------------------------------------------
    def defs(self) -> dict[str, Any]:
        out = {}
        for v in self.items:
            out.update(v.defs())
        return out

    # -------------------------------------------------------------------------
    def schema(self, version: int) -> dict[str, Any]:
        out = []
        for i in self.items:
            out.append(i.schema(version))
        return {'anyOf': out}


# =============================================================================
class Variant(Union):

    # -------------------------------------------------------------------------
    def __init__(self, data: list[dict[str, Any]], name: str, label: str):
        self.identifier = data.pop('id', None)

        if self.identifier is not None:
            name = self.identifier
            label = self.identifier

        super().__init__(data.pop('items'), name, label)

    # -------------------------------------------------------------------------
    def itemLabel(self, label: str, index: int) -> str:
        return f'{label}[{index}]'

    # -------------------------------------------------------------------------
    @property
    def objects(self) -> dict[str, Object]:
        out = super().objects
        if self.identifier is not None:
            out[self.identifier] = self
        return out

    # -------------------------------------------------------------------------
    def doc(self, since: int, until: int) -> None:
        return None

    # -------------------------------------------------------------------------
    def defs(self) -> dict[str, Any]:
        out = super().defs()

        if self.identifier is not None:
            for s in self.spans:
                n = f'{self.identifier}@{spanRef(*s)}'
                out[n] = super().schema(s[0])

        return out

    # -------------------------------------------------------------------------
    def schema(self, version: int) -> dict[str, Any]:
        if self.identifier is not None:
            s = findSpan(self.spans, version)
            if s is not None:
                return {
                    '$ref': f'#/definitions/{self.identifier}@{spanRef(*s)}',
                }
            return {}

        return super().schema(version)


# =============================================================================
class Property:

    # -------------------------------------------------------------------------
    def __init__(self, data: dict[str, Any], name: str, label: str):
        self.name = name
        self.label = label
        self.since = data.pop('since', 1)
        self.until = data.pop('until', LATEST + 1)
        self.required = data.pop('required', False)
        self.content = data.pop('sphinxDescription', data.get('description'))

        note = data.pop('sphinxNote', None)
        if note is not None:
            self.content += f'\n\n.. note::\n\n{indent(note)}'

        self._revisions = []
        for rev in data.pop('revisions', []):
            rd = deepcopy(data)
            rd.update(rev)
            self._revisions.append(Property(rd, self.name, self.label))

        if len(self._revisions):
            self.since = max(self.since,
                             min([r.since for r in self._revisions]))
            self.until = min(self.until,
                             max([r.until for r in self._revisions]))
            self.value = None
        else:
            self.value = buildValue(data, self.name, self.label)

        self.spans = None

    # -------------------------------------------------------------------------
    @property
    def objects(self) -> dict[str, Object]:
        if len(self._revisions):
            out = {}
            for r in self._revisions:
                out.update(r.objects)
            return out

        return self.value.objects

    # -------------------------------------------------------------------------
    def updateSpans(self, since: int, until: int) -> None:
        since = max(since, self.since)
        until = min(until, self.until)

        if len(self._revisions):
            spans = [[(since, until)]]
            for r in self._revisions:
                r.updateSpans(since, until)
                spans.append(r.spans)
            self.spans = combineSpans(*spans)
        else:
            self.value.updateSpans(since, until)
            self.spans = self.value.spans

    # -------------------------------------------------------------------------
    def updateRefs(self, objects: dict[str, 'Object']) -> None:
        if len(self._revisions):
            for r in self._revisions:
                r.updateRefs(objects)
        else:
            self.value.updateRefs(objects)

    # -------------------------------------------------------------------------
    def ref(self, version: int) -> dict[str, str] | None:
        s = findSpan(self.spans, version)
        if s is not None:
            return {
                '$ref': f'#/definitions/{self.label}@{spanRef(*s)}',
            }
        return None

    # -------------------------------------------------------------------------
    def doc(self, since: int, until: int) -> str | None:
        if self.content is None:
            return None

        out = f'.. _`CMakePresets.{self.label}`:\n\n``{self.name}``\n'
        if self.since > since:
            out += f'  .. presets-versionadded:: {self.since}\n\n'
        if self.until < until:
            out += f'  .. presets-versionremoved:: {self.until}\n\n'
        out += indent(self.content)

        return out

    # -------------------------------------------------------------------------
    def defs(self) -> dict[str, Any]:
        if len(self._revisions):
            out = {}
            for r in self._revisions:
                out.update(r.defs())
            return out

        return self.value.defs()

    # -------------------------------------------------------------------------
    def schema(self, version: int) -> dict[str, Any]:
        for r in self._revisions:
            if version >= r.since and version < r.until:
                return r.value.schema(version)
        return self.value.schema(version)


# =============================================================================
class CommentProperty(Property):
    name: str = '$comment'
    since: int = 10

    # -------------------------------------------------------------------------
    def __init__(self):
        data = {
            'since': type(self).since,
            'id': 'comment',
            'type': 'variant',
            'items': [
                {
                    'type': 'string',
                    'description': 'A single-line comment.',
                },
                {
                    'type': 'array',
                    'description': 'A multi-line comment.',
                    'minItems': 1,
                    'items': {
                        'type': 'string',
                        'description': 'One line of the multi-line comment.',
                    }
                }
            ],
        }
        super().__init__(data, type(self).name, type(self).name)

    # -------------------------------------------------------------------------
    def doc(self, since: int, until: int) -> None:
        return None


# -----------------------------------------------------------------------------
def buildValue(
    data: Any, name: str, label: str | None = None
) -> Variant | Object | Array | Ref | Value:
    if type(data) is dict:
        if 'anyOf' in data:
            return Union(data.pop('anyOf'), name, label)
        elif data['type'] == 'variant':
            return Variant(data, name, label)
        elif data['type'] == 'array':
            return Array(data, name, label)
        elif data['type'] == 'object':
            return Object(data, name, label)
        elif data['type'] == 'ref':
            return Ref(data)

    return Value(data)


# -----------------------------------------------------------------------------
def extractDivisions(spans: list[tuple[int, int]]) -> set[int]:
    return {n for s in spans for n in s}


# -----------------------------------------------------------------------------
def trimSpans(spans: list[tuple[int, int]],
              since: int, until: int) -> list[tuple[int, int]]:
    divisions = extractDivisions(spans)

    last = since
    out = []

    for d in sorted(divisions):
        if d > since:
            out.append((last, min(d, until)))
            last = min(d, until)
        if d >= until:
            break

    return out


# -----------------------------------------------------------------------------
def combineSpans(*spans: list[tuple[int, int]]) -> list[tuple[int, int]]:
    divisions = set()
    for s in spans:
        divisions.update(extractDivisions(s))

    divisions = sorted(divisions)
    last = divisions.pop(0)
    out = []

    for d in divisions:
        out.append((last, d))
        last = d

    return out


# -----------------------------------------------------------------------------
def findSpan(spans: list[tuple[int, int]],
             version: int) -> tuple[int, int] | None:
    for s in spans:
        if version >= s[0] and version < s[1]:
            return s

    return None


# -----------------------------------------------------------------------------
def spanRef(since: int, until: int) -> str:
    if since + 1 < until:
        if until > LATEST:
            return f'v{since}..'
        else:
            return f'v{since}..v{until - 1}'
    return f'v{since}'


# -----------------------------------------------------------------------------
def indent(text: str):
    return '\n'.join([('  ' + line).rstrip() for line in text.split('\n')])


# -----------------------------------------------------------------------------
def trimDescriptions(data: Any):
    if type(data) is dict:
        for k, v in data.items():
            if type(v) is str and k in {'description', 'sphinxDescription'}:
                data[k] = v.strip()
            elif type(v) is dict:
                data[k] = trimDescriptions(v)
            elif type(v) is list:
                data[k] = [trimDescriptions(item) for item in v]

    return data


# -----------------------------------------------------------------------------
def getPath(data: Any, path: tuple[str]) -> Any:
    if len(path) == 0:
        return data
    return getPath(data[path[0]], path[1:])


# -----------------------------------------------------------------------------
def diagnosticPresetName(symbol: str) -> str:
    sep = False
    out = ''
    for c in symbol[4:]:
        if c == '_':
            sep = True
        elif sep:
            out += c
            sep = False
        else:
            out += c.lower()
    return out


# -----------------------------------------------------------------------------
def buildDiagnosticsSchema(
    diagnostics: list[Diagnostic],
    descriptionTemplate: str,
    sphinxDescriptionTemplate: str,
) -> dict[str, dict[str, Any]]:
    out = {}

    for d in diagnostics:
        out[d.presetName] = {
            'since': d.since,
            'type': 'boolean',
            'description': d.format(descriptionTemplate),
            'sphinxDescription': d.format(sphinxDescriptionTemplate),
        }

    return out


# -----------------------------------------------------------------------------
def mergeDiagnostics(
    generated: dict[str, dict[str, Any]],
    extra: dict[str, dict[str, Any]],
) -> dict[str, dict[str, Any]]:
    # Get unsorted names.
    tail = {}
    for k in list(extra.keys()):
        s = extra[k].pop('sort', True)
        if not s:
            tail[k] = extra.pop(k)

    # Combine inputs and sort.
    out = {}
    combined = deepcopy(generated)
    combined.update(extra)
    for k in sorted(combined.keys()):
        out[k] = combined[k]

    # Add unsorted items and return result.
    out.update(tail)
    return out


# -----------------------------------------------------------------------------
def readDiagnostics(path: Path | str) -> list[Diagnostic]:
    content = ''

    # Extract diagnostics table from header.
    with open(path, 'r') as f:
        extracting = False
        for line in f:
            if DIAGNOSTIC_TABLE_MACRO in line:
                extracting = True
                continue

            if extracting:
                content += line.replace('\\\n', ' ')
                if not line.strip().endswith('\\'):
                    break

    out = []
    while True:
        m = re.match(r'\s*SELECT[(]([^)]+)[)]', content)
        if m is None:
            break

        args = [a.strip() for a in m.group(1).split(',')]
        p = diagnosticPresetName(args[3])
        c = args[3][4:].lower().replace('_', '-')
        v = int(args[4])
        if v > 1:
            out.append(Diagnostic(p, c, v))

        content = content[m.span()[1]:]

    return out


# -----------------------------------------------------------------------------
def main():
    # Read the schema definition.
    with open(PRESETS / SCHEMA_YAML_FILENAME, 'r') as f:
        schema = yaml.safe_load(f)

    # Extract the current (latest) version.
    global LATEST

    del schema['definitions']
    LATEST = schema.pop('version')

    # Remove whitespace around descriptions.
    schema = trimDescriptions(schema)

    # Read diagnostics and update schema.
    diagnostics = readDiagnostics(DIAGNOSTICS)
    configureSchema = getPath(schema, CONFIGURE_PRESET_PROPERTIES_PATH)
    configureSchema['warnings']['properties'] = mergeDiagnostics(
        buildDiagnosticsSchema(diagnostics, WARNING_DESCRIPTION,
                               WARNING_SPHINX_DESCRIPTION),
        configureSchema['warnings']['properties'])
    configureSchema['errors']['properties'] = mergeDiagnostics(
        buildDiagnosticsSchema(diagnostics, ERROR_DESCRIPTION,
                               ERROR_SPHINX_DESCRIPTION),
        configureSchema['errors']['properties'])

    # Extract global type definitions.
    global TYPES, REFS

    for t in schema.pop('types', []):
        name = t['id']
        value = buildValue(t, name)
        value.updateSpans(1, LATEST + 1)
        TYPES[name] = value

    for t in TYPES.values():
        t.updateRefs(TYPES)

    # Reset the spans for references, as these currently contain only spans for
    # circular references of global types, which are likely broader than the
    # actual usage spans of those types. We'll recalculate the correct spans
    # when we update refs from the root object.
    REFS = {}

    # Parse the root object. This will also recursively parse other objects.
    root = Object(schema, 'root')
    root.updateSpans(1, LATEST + 1)

    # Resolve references.
    objects = root.objects

    types = objects
    types.update(TYPES)

    root.updateRefs(types)

    for name, value in TYPES.items():
        span = REFS.get(name)
        if span is not None:
            value.updateSpans(*span)

    root.updateSpans(1, LATEST + 1)

    # Extract Sphinx documentation for each object's properties.
    for n, o in objects.items():
        doc = o.doc(1, LATEST + 1)
        if doc is not None:
            print(f'- Generating reST documentation for {n} properties')
            with open(PRESETS / f'{o.name}-properties.rst', 'w') as f:
                print(RST_BANNER, file=f)
                print(doc, file=f)

    # Generate schema for each version.
    versions = []
    for v in range(1, LATEST + 1):
        vs = root.schema(v, inline={'version'})
        vs['properties']['version']['const'] = v
        versions.append(vs)

    # Add definitions and write the schema.
    types = root.defs(exclude={'version'})
    for t in TYPES.values():
        types.update(t.defs())

    schema = {
        '$schema': 'http://json-schema.org/draft/2020-12/schema#',
        'type': 'object',
        'oneOf': versions,
        'required': ['version'],
        'definitions': types
    }

    with open(PRESETS / SCHEMA_JSON_FILENAME, 'w') as f:
        print(json.dumps(schema, indent=2), file=f)


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

main()



import json
import jsonschema
import sys

with open(sys.argv[1], "r", encoding="utf-8-sig") as f:
    contents = json.load(f)

with open(sys.argv[2], "r", encoding="utf-8") as f:
    schema = json.load(f)

try:
    jsonschema.validate(contents, schema)
except jsonschema.ValidationError as e:
    print(e)
    sys.exit(1)
